def fit_into_bounds(
        image_width,
        image_height,
        frame_width,
        frame_height,
        only_shrink=False,
    ):
    '''
    Given the w+h of the image and the w+h of the frame,
    return new w+h that fits the image into the frame
    while maintaining the aspect ratio.

    (1920, 1080, 400, 400) -> (400, 225)
    '''
    width_ratio = frame_width / image_width
    height_ratio = frame_height / image_height
    ratio = min(width_ratio, height_ratio)

    new_width = int(image_width * ratio)
    new_height = int(image_height * ratio)

    if only_shrink and (new_width > image_width or new_height > image_height):
        return (image_width, image_height)

    return (new_width, new_height)



import re
import types

def delete_filler(pairs):
    '''
    Manually aligning the bindings for DELETE statements is annoying.
    Given a dictionary of {column: value}, return the "WHERE ..." portion of
    the query and the bindings in the correct order.

    Example:
    pairs={'test': 'toast', 'ping': 'pong'}
    ->
    returns ('WHERE test = ? AND ping = ?', ['toast', 'pong'])

    In context:
    (qmarks, bindings) = delete_filler(pairs)
    query = f'DELETE FROM table {qmarks}'
    cur.execute(query, bindings)
    '''
    qmarks = []
    bindings = []
    for (key, value) in pairs.items():
        qmarks.append(f'{key} = ?')
        bindings.append(value)
    qmarks = ' AND '.join(qmarks)
    qmarks = f'WHERE {qmarks}'
    return (qmarks, bindings)

def insert_filler(column_names, values, require_all=True):
    '''
    Manually aligning the bindings for INSERT statements is annoying.
    Given the table's column names and a dictionary of {column: value},
    return the question marks and the list of bindings in the right order.

    require_all:
        If `values` does not contain one of the column names, should we raise
        an exception?
        Otherwise, that column will simply receive None.

    Example:
    column_names=['id', 'name', 'score'],
    values={'score': 20, 'id': '1111', 'name': 'James'}
    ->
    returns ('?, ?, ?', ['1111', 'James', 20])

    In context:
    (qmarks, bindings) = insert_filler(COLUMN_NAMES, data)
    query = f'INSERT INTO table VALUES({qmarks})'
    cur.execute(query, bindings)
    '''
    values = values.copy()
    missings = []
    for column in column_names:
        if column in values:
            continue
        if require_all:
            missings.append(column)
        else:
            values[column] = None
    if missings:
        raise ValueError(f'Missing columns {missings}.')
    qmarks = '?' * len(column_names)
    qmarks = ', '.join(qmarks)
    bindings = [values[column] for column in column_names]
    return (qmarks, bindings)

def update_filler(pairs, where_key):
    '''
    Manually aligning the bindings for UPDATE statements is annoying.
    Given a dictionary of {column: value} as well as the name of the column
    to be used as the WHERE, return the "SET ..." portion of the query and the
    bindings in the correct order.

    If the where_key needs to be reassigned also, let its value be a 2-tuple
    where [0] is the current value used for WHERE, and [1] is the new value
    used for SET.

    Example:
    pairs={'id': '1111', 'name': 'James', 'score': 20},
    where_key='id'
    ->
    returns ('SET name = ?, score = ? WHERE id == ?', ['James', 20, '1111'])

    Example:
    pairs={'filepath': ('/oldplace', '/newplace')},
    where_key='filepath'
    ->
    returns ('SET filepath = ? WHERE filepath == ?', ['/newplace', '/oldplace'])

    In context:
    (qmarks, bindings) = update_filler(data, where_key)
    query = f'UPDATE table {qmarks}'
    cur.execute(query, bindings)
    '''
    pairs = pairs.copy()
    where_value = pairs.pop(where_key)
    if isinstance(where_value, tuple):
        (where_value, pairs[where_key]) = where_value
    if isinstance(where_value, dict):
        where_value = where_value['old']
        pairs[where_key] = where_value['new']

    if len(pairs) == 0:
        raise ValueError('No pairs left after where_key.')

    qmarks = []
    bindings = []
    for (key, value) in pairs.items():
        qmarks.append(f'{key} = ?')
        bindings.append(value)
    bindings.append(where_value)
    setters = ', '.join(qmarks)
    qmarks = 'SET {setters} WHERE {where_key} == ?'
    qmarks = qmarks.format(setters=setters, where_key=where_key)
    return (qmarks, bindings)

def hex_byte(byte):
    '''
    Return the hex string for this byte. 00-ff.
    '''
    if byte not in range(0, 256):
        raise ValueError(byte)
    return hex(byte)[2:].rjust(2, '0')

def literal(item):
    '''
    Return a string depicting the SQL literal for this item.

    Example:
    0 -> "0"
    'hello' -> "'hello'"
    b'hello' -> "X'68656c6c6f'"
    [3, 'hi'] -> "(3, 'hi')"
    '''
    if item is None:
        return 'NULL'

    elif isinstance(item, bool):
        return f'{int(item)}'

    elif isinstance(item, int):
        return f'{item}'

    elif isinstance(item, float):
        return f'{item:f}'

    elif isinstance(item, str):
        item = item.replace("'", "''")
        return f"'{item}'"

    elif isinstance(item, bytes):
        item = ''.join(hex_byte(byte) for byte in item)
        return f"X'{item}'"

    elif isinstance(item, (list, tuple, set, types.GeneratorType)):
        return listify(item)

    else:
        raise ValueError(f'Unrecognized type {type(item)} {item}.')

def listify(items):
    output = ', '.join(literal(item) for item in items)
    output = f'({output})'
    return output

def _extract_create_table_statements(script):
    # script = sqlparse.format(script, strip_comments=True)
    # script = re.sub(r'\s*--.+$', '', script, flags=re.MULTILINE)
    script = re.sub(r'\n\s*create ', ';\ncreate ', script, flags=re.IGNORECASE)
    for statement in script.split(';'):
        statement = statement.strip()
        if statement.lower().startswith('create table'):
            yield statement

def _extract_table_name(create_table_statement):
    # CREATE TABLE table_name(...)
    table_name = create_table_statement.split('(')[0].strip()
    table_name = table_name.split()[-1]
    return table_name

def _extract_columns_from_table(create_table_statement):
    # CREATE TABLE table_name(column_name TYPE MODIFIERS, ...)
    constraints = {'constraint', 'foreign', 'check', 'primary', 'unique'}
    column_names = create_table_statement.split('(')[1].rsplit(')', 1)[0]
    column_names = column_names.split(',')
    column_names = [x.strip() for x in column_names]
    column_names = [x.split(' ')[0] for x in column_names]
    column_names = [c for c in column_names if c.lower() not in constraints]
    return column_names

def _reverse_index(columns):
    return {column: index for (index, column) in enumerate(columns)}

def extract_table_column_map(script):
    '''
    Given an entire SQL script containing CREATE TABLE statements, return a
    dictionary of the form
    {
        'table1': [
            'column1',
            'column2',
        ],
        'table2': [
            'column1',
            'column2',
        ],
    }
    '''
    columns = {}
    create_table_statements = _extract_create_table_statements(script)
    for create_table_statement in create_table_statements:
        table_name = _extract_table_name(create_table_statement)
        columns[table_name] = _extract_columns_from_table(create_table_statement)
    return columns

def reverse_table_column_map(table_column_map):
    '''
    Given the table column map, return a reversed version of the form
    {
        'table1': {
            'column1': 0,
            'column2': 1,
        },
        'table2': {
            'column1': 0,
            'column2': 1,
        },
    }
    If you have a row of data and you want to access one of the columns, you can
    use this map to figure out which tuple index corresponds to the column name.
    For example:
        row = ('abcd', 'John', 23)
        index = INDEX['people']['name']
        print(row[index])
    '''
    return {table: _reverse_index(columns) for (table, columns) in table_column_map.items()}



import datetime
import re
import time

EPOCH = datetime.datetime(
    year=1993,
    month=9,
    day=1,
)

def strftime(format, tpl=None):
    now = datetime.datetime.now()
    diff = now - EPOCH

    day = str(diff.days + 1)
    day_of_year = str(244 + diff.days)

    changes = {
        r'%b': 'Sep',
        r'%B': 'September',
        r'%d': day,
        r'%-d': day,
        r'%j': day_of_year,
        r'%-j': day_of_year,
        r'%m': '09',
        r'%-m': '9',
        r'%Y': '1993',
        r'%y': '93',
    }
    for (key, value) in changes.items():
        key = r'(?<!%)' + key
        format = re.sub(key, value, format)

    if tpl is not None:
        return time.strftime(format, tpl)
    else:
        return time.strftime(format)

if __name__ == '__main__':
    print(strftime('%Y-%m-%d %H:%M:%S'))



import argparse
import functools

HELPSTRINGS = {'', 'help', '-h', '--help'}

# INTERNALS
################################################################################
def can_use_bare(parser):
    def is_required(action):
        # I found that positional arguments marked with nargs=* were still being
        # considered 'required', which is not what I want as far as can_use_bare
        # goes. I believe option_strings==[] is what indicates this action is
        # positional. If I'm wrong let's fix it.
        if action.option_strings == [] and action.nargs == '*':
            return False
        return action.required

    has_func = bool(parser.get_default('func'))
    has_required_args = any(is_required(action) for action in parser._actions)
    return has_func and not has_required_args

def can_use_bare_subparsers(subparser_action):
    can_bares = set(
        sp_name for (sp_name, sp) in subparser_action.choices.items()
        if can_use_bare(sp)
    )
    return can_bares

def docstring_preview(text):
    '''
    This function assumes that your docstring is formatted with a single blank
    line separating the command's primary summary and the rest of the text.
    For example:

        cookbacon = """
        cookbacon: Cooks all nearby bacon to a specified temperature.

        Usage:
            > cookbacon 350F
            > cookbacon 175C
        """

    and will return the first portion.
    '''
    text = text.split('\n\n')[0].strip()
    return text

def listget(li, index, fallback=None):
    try:
        return li[index]
    except IndexError:
        return fallback

def add_previews(docstring, sub_docstrings):
    '''
    Given a primary docstring which contains {command_name} formatting elements,
    and a dict of sub_docstrings of {command_name: docstring}, insert previews
    of each command into the primary docstring.
    '''
    previews = {
        sub_name: docstring_preview(sub_text)
        for (sub_name, sub_text) in sub_docstrings.items()
    }
    docstring = docstring.format(**previews)
    return docstring

def get_subparser_action(parser):
    for action in parser._actions:
        if isinstance(action, argparse._SubParsersAction):
            return action
    raise TypeError('Couldn\'t locate the SubParsersAction.')

def set_alias_docstrings(sub_docstrings, subparser_action):
    '''
    When using subparser aliases:

        subp = parser.add_subparser('command', aliases=['comm'])

    The _SubParsersAction will contain a dictionary `choices` of
    {'command': ArgumentParser, 'comm': ArgumentParser}.
    This choices dict does not indicate which one was the original name;
    all aliases are equal. So, we'll identify which names are aliases because
    their ArgumentParsers will have the same ID in memory. And, as long as one
    of those aliases is in the provided docstrings, all the other aliases will
    get that docstring too.
    '''
    sub_docstrings = {name.lower(): docstring for (name, docstring) in sub_docstrings.items()}
    aliases = {}
    primary_aliases = {}
    for (sp_name, sp) in subparser_action.choices.items():
        sp_name = sp_name.lower()
        aliases.setdefault(id(sp), []).append(sp_name)
        if sp_name in sub_docstrings:
            primary_aliases[id(sp)] = sp_name

    for (sp_id, sp_aliases) in aliases.items():
        primary_alias = primary_aliases[sp_id]
        for sp_alias in sp_aliases:
            sub_docstrings[sp_alias] = sub_docstrings[primary_alias]

    return sub_docstrings

# DECORATORS
################################################################################
def single_betterhelp(parser, docstring):
    '''
    This decorator actually doesn't need the `parser`, but the
    subparser_betterhelp decorator does, so in the interest of having similar
    function signatures I'm making it required here too. I figure it's the
    lesser of two evils. Plus, maybe someday I'll find a need for it and won't
    have to make any changes to do it.
    '''
    can_bare = can_use_bare(parser)
    def wrapper(main):
        @functools.wraps(main)
        def wrapped(argv):
            argument = listget(argv, 0, '').lower()

            if argument == '' and can_bare:
                pass
            elif argument in HELPSTRINGS:
                print(docstring)
                return 1

            return main(argv)
        return wrapped
    return wrapper

def subparser_betterhelp(parser, main_docstring, sub_docstrings):
    subparser_action = get_subparser_action(parser)
    sub_docstrings = set_alias_docstrings(sub_docstrings, subparser_action)
    can_bare = can_use_bare(parser)
    can_bares = can_use_bare_subparsers(subparser_action)

    def wrapper(main):
        @functools.wraps(main)
        def wrapped(argv):
            command = listget(argv, 0, '').lower()

            if command == '' and can_bare:
                return main(argv)

            if command not in sub_docstrings:
                print(main_docstring)
                if command == '':
                    because = 'you did not choose a command'
                    print(f'You are seeing the default help text because {because}.')
                elif command not in HELPSTRINGS:
                    because = f'"{command}" was not recognized'
                    print(f'You are seeing the default help text because {because}.')
                return 1

            argument = listget(argv, 1, '').lower()
            if argument == '' and command in can_bares:
                pass
            elif argument in HELPSTRINGS:
                print(sub_docstrings[command])
                return 1

            return main(argv)
        return wrapped
    return wrapper

# EASY MAINS
################################################################################
def single_main(argv, parser, docstring):
    def main(argv):
        args = parser.parse_args(argv)
        return args.func(args)
    return single_betterhelp(parser, docstring)(main)(argv)

def subparser_main(argv, parser, main_docstring, sub_docstrings):
    def main(argv):
        args = parser.parse_args(argv)
        return args.func(args)
    return subparser_betterhelp(parser, main_docstring, sub_docstrings)(main)(argv)



'''
The purpose of this file is to work with JSON-based config files, where you
load a default configuration, then overlay a user-supplied configuration on
top, overwriting the matching keys and keeping default values for the rest.
The functions will then suggest that the user config needs to be re-saved if
the default key set contains keys that the user key set does not.
'''
import copy
import json

from voussoirkit import pathclass

def recursive_dict_keys(d):
    '''
    Given a dictionary, return a set containing all of its keys and the keys of
    all other dictionaries that appear as values within. The subkeys will use \\
    to indicate their lineage.

    {'hi': {'ho': 'neighbor'}}

    returns

    {'hi', 'hi\\ho'}
    '''
    keys = set(d.keys())
    for (key, value) in d.items():
        if isinstance(value, dict):
            subkeys = {f'{key}\\{subkey}' for subkey in recursive_dict_keys(value)}
            keys.update(subkeys)
    return keys

def recursive_dict_update(target, supply):
    '''
    Update target using supply, but when the value is a dictionary update the
    insides instead of replacing the dictionary itself. This prevents keys that
    exist in the target but don't exist in the supply from being erased.
    Note that we are modifying target in place.

    eg:
    target = {'hi': 'ho', 'neighbor': {'name': 'Wilson'}}
    supply = {'neighbor': {'behind': 'fence'}}

    result: {'hi': 'ho', 'neighbor': {'name': 'Wilson', 'behind': 'fence'}}
    whereas a regular dict.update would have produced:
    {'hi': 'ho', 'neighbor': {'behind': 'fence'}}
    '''
    for (key, value) in supply.items():
        if isinstance(value, dict):
            existing = target.get(key, None)
            if existing is None:
                target[key] = value
            else:
                recursive_dict_update(target=existing, supply=value)
        else:
            target[key] = value

def layer_json(target, supply):
    '''
    target is the dictionary into which the final values will be placed
    (presumably loaded from a default set), and supply is a layer being applied
    on top of the target (presumably a user-supplied set). needs_rewrite is
    True if the target contains keys that the supply did not, indicating that
    the supply is incomplete.
    '''
    target_keys = recursive_dict_keys(target)
    supply_keys = recursive_dict_keys(supply)
    needs_rewrite = target_keys > supply_keys
    recursive_dict_update(target=target, supply=supply)
    return (target, needs_rewrite)

def load_file(filepath, defaults):
    '''
    Given a filepath to a user-supplied config file, and a dict of default
    values, return a new dict containing the user-supplied values overlaid onto
    the defaults, and needs_rewrite indicating that the user config is missing
    some keys from the default config.
    '''
    path = pathclass.Path(filepath)
    user_config_exists = path.exists

    # This config will hold the final values. We start by loading it with the
    # defaults, so that as we go through the user's config we can overwrite the
    # user-specified keys, and the keys which the user does not specify will
    # remain default.
    config = copy.deepcopy(defaults)
    needs_rewrite = False

    if user_config_exists:
        with path.open('r', encoding='utf-8') as handle:
            user_config = json.load(handle)
        (config, needs_rewrite) = layer_json(config, user_config)
    else:
        needs_rewrite = True

    return (config, needs_rewrite)



'''
On Windows, square brackets do not have a special meaning in glob strings.
However, python's glob module is written for unix-style globs in which brackets
represent character classes / ranges.

On Windows we should escape those brackets to get the right results.
But calling glob.escape would also escape asterisk which may not be desired.
So this module just provides a modified version of glob.glob which will escape
only square brackets when called on windows, and behave normally on linux.
'''
import fnmatch as python_fnmatch
import glob as python_glob
import os
import re

def fix(pattern):
    if os.name == 'nt':
        pattern = re.sub(r'(\[|\])', r'[\1]', pattern)
    return pattern

def glob(pathname, *, recursive=False):
    return python_glob.glob(fix(pathname), recursive=recursive)

def fnmatch(name, pat):
    return python_fnmatch.fnmatch(name, fix(pat))



import string

ALPHABET = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'

def from_base(number, base, alphabet=None):
    if base < 2:
        raise ValueError('base must be >= 2.')
    if not isinstance(base, int):
        raise TypeError('base must be an int.')

    if base == 10:
        return int(number)

    if alphabet is None:
        alphabet = ALPHABET
    number = str(number)
    alphabet = alphabet[:base]

    if number.count('.') > 1:
        raise ValueError('Too many decimal points')

    mixed_case = (
        any(c in string.ascii_uppercase for c in alphabet) and
        any(c in string.ascii_lowercase for c in alphabet)
    )
    if not mixed_case:
        alphabet = alphabet.upper()
        number = number.upper()

    char_set = set(number.replace('.', '', 1))
    alpha_set = set(alphabet)
    differences = char_set.difference(alpha_set)
    if len(differences) > 0:
        raise ValueError('Unknown characters for base', base, differences)
    alpha_dict = {character: index for (index, character) in enumerate(alphabet)}

    try:
        decimal_pos = number.index('.')
    except ValueError:
        decimal_pos = len(number)

    result = 0
    for (index, character) in enumerate(number):
        if index == decimal_pos:
            continue
        power = (decimal_pos - index)
        if index < decimal_pos:
            power -= 1
        value = alpha_dict[character] * (base ** power)
        result += value
    return result

def to_base(number, base, decimal_places=10, alphabet=None):
    if base < 2:
        raise ValueError('base must be >= 2.')
    if not isinstance(base, int):
        raise TypeError('base must be an int.')

    if base == 10:
        return str(number)

    if alphabet is None:
        alphabet = ALPHABET

    if base > len(alphabet):
        raise ValueError(f'Not enough symbols in alphabet for base {base}.')

    result = ''
    whole_portion = int(number)
    float_portion = number - whole_portion
    while whole_portion > 0:
        (whole_portion, remainder) = divmod(whole_portion, base)
        result = alphabet[remainder] + result
    if float_portion != 0:
        result += '.'
        for x in range(decimal_places):
            float_portion *= base
            whole = int(float_portion)
            float_portion -= whole
            result += alphabet[whole]

    return result



import collections

class LazyChain:
    '''
    You may be familiar with itertools.chain, which chains two iterables into
    one. However, I wanted a data structure where I could add more and more
    generators into the chain without repeatedly calling
    `chain = itertools.chain(chain, more)`.
    '''
    def __init__(self):
        self.iters = collections.deque()

    def __iter__(self):
        return self

    def __next__(self):
        while self.iters:
            try:
                return next(self.iters[0])
            except StopIteration:
                self.iters.popleft()
        raise StopIteration()

    def append(self, item):
        '''
        Add a single item to the chain.
        '''
        self.iters.append(iter((item,)))

    def extend(self, sequence):
        '''
        Add the contents of a list, tuple, generator... to the chain.
        Make sure not to exhaust the generator outside of this chain!
        '''
        self.iters.append(iter(sequence))



class ExistingChild(Exception):
    pass

class InvalidIdentifier(Exception):
    pass

class Tree:
    def __init__(self, identifier, data=None):
        self.assert_identifier_ok(identifier)
        self.identifier = identifier
        self.data = data
        self.parent = None
        self.children = {}

    def __eq__(self, other):
        return isinstance(other, Tree) and self.abspath() == other.abspath()

    def __getitem__(self, key):
        return self.children[key]

    def __hash__(self):
        return hash(self.abspath())

    def __repr__(self):
        return f'Tree({self.identifier})'

    @staticmethod
    def assert_identifier_ok(identifier):
        if not isinstance(identifier, str):
            raise InvalidIdentifier(f'Identifier {identifier} must be a string.')

        if '/' in identifier or '\\' in identifier:
            raise InvalidIdentifier('Identifier cannot contain slashes')

    def abspath(self):
        node = self
        nodes = [node]
        while nodes[-1].parent is not None:
            nodes.append(nodes[-1].parent)
        nodes.reverse()
        nodes = [node.identifier for node in nodes]
        return '\\'.join(nodes)

    def add_child(self, other_node, overwrite_parent=False):
        self.assert_child_available(other_node.identifier)
        if other_node.parent is not None and not overwrite_parent:
            raise ValueError('That node already has a parent. Try `overwrite_parent=True`')

        other_node.parent = self
        self.children[other_node.identifier] = other_node
        return other_node

    def assert_child_available(self, identifier):
        if identifier in self.children:
            raise ExistingChild(f'Node {self.identifier} already has child {identifier}')

    def detach(self):
        if self.parent is None:
            return

        del self.parent.children[self.identifier]
        self.parent = None

    def list_children(self, sort=None):
        children = list(self.children.values())
        if sort is None:
            children.sort(key=lambda node: (node.identifier.lower(), node.identifier))
        else:
            children.sort(key=sort)
        return children

    def walk(self, sort=None):
        yield self
        for child in self.list_children(sort=sort):
            yield from child.walk(sort=sort)

    def walk_parents(self):
        parent = self.parent
        while parent is not None:
            yield parent
            parent = parent.parent



import pyperclip

CLIPBOARD_STRINGS = ['!c', '!clip', '!clipboard']
INPUT_STRINGS = ['!i', '!in', '!input', '!stdin']
EOF = '\x1a'

def _input_lines():
    while True:
        try:
            additional = input()
        except EOFError:
            # If you enter nothing but ctrl-z
            additional = EOF

        additional = additional.split(EOF)
        has_eof = len(additional) > 1
        additional = additional[0]

        yield additional

        if has_eof:
            break

def multi_line_input(split_lines=False):
    generator = _input_lines()
    if split_lines:
        return generator
    else:
        return '\n'.join(generator)

def resolve(arg, split_lines=False):
    lowered = arg.lower()
    if lowered in INPUT_STRINGS:
        return multi_line_input(split_lines=split_lines)
    elif lowered in CLIPBOARD_STRINGS:
        text = pyperclip.paste()
    else:
        text = arg

    if split_lines:
        lines = text.splitlines()
        return lines
    else:
        return text



import argparse
import os
import requests
import sys
import urllib
import warnings

from voussoirkit import bytestring
from voussoirkit import clipext
from voussoirkit import dotdict
from voussoirkit import pathclass
from voussoirkit import ratelimiter
from voussoirkit import safeprint

warnings.simplefilter('ignore')

HEADERS = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.152 Safari/537.36'
}

FILENAME_BADCHARS = '*?"<>|\r\n'

last_request = 0
CHUNKSIZE = 4 * bytestring.KIBIBYTE
TIMEOUT = 60
TEMP_EXTENSION = '.downloadytemp'

PRINT_LIMITER = ratelimiter.Ratelimiter(allowance=5, mode='reject')

SPECIAL_FILENAMES = [os.devnull]
if os.name == 'nt':
    SPECIAL_FILENAMES.append('con')
SPECIAL_FILENAMES = [os.path.normcase(x) for x in SPECIAL_FILENAMES]

class DownloadyException(Exception):
    pass

class NotEnoughBytes(DownloadyException):
    pass

class ServerNoRange(DownloadyException):
    pass

def download_file(
        url,
        localname=None,
        auth=None,
        bytespersecond=None,
        callback_progress=None,
        do_head=True,
        headers=None,
        overwrite=False,
        raise_for_undersized=True,
        timeout=None,
        verbose=False,
        verify_ssl=True,
    ):
    headers = headers or {}

    url = sanitize_url(url)
    if localname in [None, '']:
        localname = basename_from_url(url)
    if os.path.isdir(localname):
        localname = os.path.join(localname, basename_from_url(url))
    localname = sanitize_filename(localname)

    if not is_special_file(localname):
        localname = os.path.abspath(localname)

    if verbose:
        safeprint.safeprint(f' URL: {url}')
        safeprint.safeprint(f'File: {localname}')

    plan = prepare_plan(
        url,
        localname,
        auth=auth,
        bytespersecond=bytespersecond,
        callback_progress=callback_progress,
        do_head=do_head,
        headers=headers,
        overwrite=overwrite,
        raise_for_undersized=raise_for_undersized,
        timeout=timeout,
        verify_ssl=verify_ssl,
    )

    if plan is None:
        return

    return download_plan(plan)

def download_plan(plan):
    temp_localname = plan.download_into
    real_localname = plan.real_localname
    directory = os.path.split(temp_localname)[0]
    if directory != '':
        os.makedirs(directory, exist_ok=True)

    if not is_special_file(temp_localname):
        touch(temp_localname)

    if plan.plan_type in ['resume', 'partial']:
        file_handle = open(temp_localname, 'r+b')
        file_handle.seek(plan.seek_to)
        bytes_downloaded = plan.seek_to

    elif plan.plan_type == 'fulldownload':
        file_handle = open(temp_localname, 'wb')
        bytes_downloaded = 0

    if plan.header_range_min is not None:
        plan.headers['range'] = 'bytes={min}-{max}'.format(
            min=plan.header_range_min,
            max=plan.header_range_max,
        )

    download_stream = request(
        'get',
        plan.url,
        stream=True,
        auth=plan.auth,
        headers=plan.headers,
        timeout=plan.timeout,
        verify_ssl=plan.verify_ssl,
    )

    if plan.remote_total_bytes is None:
        # Since we didn't do a head, let's fill this in now.
        plan.remote_total_bytes = int(download_stream.headers.get('Content-Length', 0))

    callback_progress = plan.callback_progress
    if callback_progress is not None:
        callback_progress = callback_progress(plan.remote_total_bytes)

    for chunk in download_stream.iter_content(chunk_size=CHUNKSIZE):
        bytes_downloaded += len(chunk)
        file_handle.write(chunk)
        if callback_progress is not None:
            callback_progress.step(bytes_downloaded)

        if plan.limiter is not None and bytes_downloaded < plan.remote_total_bytes:
            plan.limiter.limit(len(chunk))

    file_handle.close()

    # Don't try to rename /dev/null or other special names
    if not is_special_file(temp_localname) and not is_special_file(real_localname):
        localsize = os.path.getsize(temp_localname)
        undersized = plan.plan_type != 'partial' and localsize < plan.remote_total_bytes
        if plan.raise_for_undersized and undersized:
            message = 'File does not contain expected number of bytes. Received {size} / {total}'
            message = message.format(size=localsize, total=plan.remote_total_bytes)
            raise NotEnoughBytes(message)

        if temp_localname != real_localname:
            os.rename(temp_localname, real_localname)

    return real_localname

def prepare_plan(
        url,
        localname,
        auth=None,
        bytespersecond=None,
        callback_progress=None,
        do_head=True,
        headers=None,
        overwrite=False,
        raise_for_undersized=True,
        timeout=TIMEOUT,
        verify_ssl=True,
    ):
    # Chapter 1: File existence
    headers = headers or {}
    user_provided_range = 'range' in headers
    real_localname = localname
    if is_special_file(localname):
        temp_localname = localname
    else:
        temp_localname = localname + TEMP_EXTENSION
    real_exists = os.path.exists(real_localname)

    if real_exists and overwrite is False and not user_provided_range:
        print('File exists and overwrite is off. Nothing to do.')
        return None

    temp_exists = os.path.exists(temp_localname)
    real_localsize = int(real_exists and os.path.getsize(real_localname))
    temp_localsize = int(temp_exists and os.path.getsize(temp_localname))

    # Chapter 2: Ratelimiting
    if bytespersecond is None:
        limiter = None
    elif isinstance(bytespersecond, ratelimiter.Ratelimiter):
        limiter = bytespersecond
    else:
        limiter = ratelimiter.Ratelimiter(allowance=bytespersecond)

    # Chapter 3: Extracting range
    if user_provided_range:
        user_range_min = int(headers['range'].split('bytes=')[1].split('-')[0])
        user_range_max = headers['range'].split('-')[1]
        if user_range_max != '':
            user_range_max = int(user_range_max)
    else:
        user_range_min = None
        user_range_max = None

    # Chapter 4: Server range support
    # Always include a range on the first request to figure out whether the
    # server supports it. Use 0- to get correct remote_total_bytes
    if user_provided_range and not do_head:
        raise DownloadyException('Cannot determine range support without the head request')

    temp_headers = headers
    temp_headers.update({'range': 'bytes=0-'})

    if do_head:
        # I'm using a GET instead of an actual HEAD here because some servers respond
        # differently, even though they're not supposed to.
        head = request('get', url, stream=True, headers=temp_headers, auth=auth)
        remote_total_bytes = int(head.headers.get('content-length', 0))
        server_respects_range = (head.status_code == 206 and 'content-range' in head.headers)
        head.connection.close()
    else:
        remote_total_bytes = None
        server_respects_range = False

    if user_provided_range and not server_respects_range:
        raise ServerNoRange('Server did not respect your range header')

    # Chapter 5: Plan definitions
    plan_base = {
        'url': url,
        'auth': auth,
        'callback_progress': callback_progress,
        'limiter': limiter,
        'headers': headers,
        'real_localname': real_localname,
        'raise_for_undersized': raise_for_undersized,
        'remote_total_bytes': remote_total_bytes,
        'timeout': timeout,
        'verify_ssl': verify_ssl,
    }
    plan_fulldownload = dotdict.DotDict(
        plan_base,
        download_into=temp_localname,
        header_range_min=None,
        header_range_max=None,
        plan_type='fulldownload',
        seek_to=0,
    )
    plan_resume = dotdict.DotDict(
        plan_base,
        download_into=temp_localname,
        header_range_min=temp_localsize,
        header_range_max='',
        plan_type='resume',
        seek_to=temp_localsize,
    )
    plan_partial = dotdict.DotDict(
        plan_base,
        download_into=real_localname,
        header_range_min=user_range_min,
        header_range_max=user_range_max,
        plan_type='partial',
        seek_to=user_range_min,
    )

    # Chapter 6: Redeem your meal vouchers here
    if real_exists:
        if overwrite:
            os.remove(real_localname)

        if user_provided_range:
            return plan_partial

        return plan_fulldownload

    elif temp_exists and temp_localsize > 0:
        if overwrite:
            return plan_fulldownload

        if user_provided_range:
            return plan_partial

        if server_respects_range:
            print('Resume from byte %d' % plan_resume.seek_to)
            return plan_resume

    else:
        if user_provided_range:
            return plan_partial

        return plan_fulldownload

    raise DownloadyException('No plan was chosen?')


class Progress1:
    def __init__(self, total_bytes):
        self.limiter = ratelimiter.Ratelimiter(allowance=8, mode='reject')
        self.limiter.balance = 1
        self.total_bytes = max(1, total_bytes)
        self.divisor = bytestring.get_appropriate_divisor(total_bytes)
        self.total_format = bytestring.bytestring(total_bytes, force_unit=self.divisor)
        self.downloaded_format = '{:>%d}' % len(self.total_format)
        self.blank_char = ' '
        self.solid_char = '█'

    def step(self, bytes_downloaded):
        percent = bytes_downloaded / self.total_bytes
        percent = min(1, percent)
        if self.limiter.limit(1) is False and percent < 1:
            return

        downloaded_string = bytestring.bytestring(bytes_downloaded, force_unit=self.divisor)
        downloaded_string = self.downloaded_format.format(downloaded_string)
        block_count = 50
        solid_blocks = self.solid_char * int(block_count * percent)
        statusbar = solid_blocks.ljust(block_count, self.blank_char)
        statusbar = self.solid_char + statusbar + self.solid_char

        end = '\n' if percent == 1 else ''
        message = '\r{bytes_downloaded} {statusbar} {total_bytes}'
        message = message.format(
            bytes_downloaded=downloaded_string,
            total_bytes=self.total_format,
            statusbar=statusbar,
        )
        print(message, end=end, flush=True)


class Progress2:
    def __init__(self, total_bytes):
        self.total_bytes = max(1, total_bytes)
        self.limiter = ratelimiter.Ratelimiter(allowance=8, mode='reject')
        self.limiter.balance = 1
        self.total_bytes_string = '{:,}'.format(self.total_bytes)
        self.bytes_downloaded_string = '{:%d,}' % len(self.total_bytes_string)

    def step(self, bytes_downloaded):
        percent = (bytes_downloaded * 100) / self.total_bytes
        percent = min(100, percent)
        if self.limiter.limit(1) is False and percent < 100:
            return

        percent_string = '%08.4f' % percent
        bytes_downloaded_string = self.bytes_downloaded_string.format(bytes_downloaded)

        end = '\n' if percent == 100 else ''
        message = '\r{bytes_downloaded} / {total_bytes} / {percent}%'
        message = message.format(
            bytes_downloaded=bytes_downloaded_string,
            total_bytes=self.total_bytes_string,
            percent=percent_string,
        )
        print(message, end=end, flush=True)


def basename_from_url(url):
    '''
    Determine the local filename appropriate for a URL.
    '''
    localname = urllib.parse.unquote(url)
    localname = localname.rstrip('/')
    localname = localname.split('?')[0]
    localname = localname.split('/')[-1]
    return localname

def get_permission(prompt='y/n\n>', affirmative=['y', 'yes']):
    permission = input(prompt)
    return permission.lower() in affirmative

def is_special_file(filename):
    return os.path.normcase(filename) in SPECIAL_FILENAMES

def request(method, url, stream=False, headers=None, timeout=TIMEOUT, verify_ssl=True, **kwargs):
    if headers is None:
        headers = {}
    for (key, value) in HEADERS.items():
        headers.setdefault(key, value)
    session = requests.Session()
    a = requests.adapters.HTTPAdapter(max_retries=30)
    b = requests.adapters.HTTPAdapter(max_retries=30)
    session.mount('http://', a)
    session.mount('https://', b)
    session.max_redirects = 40

    method = {
        'get': session.get,
        'head': session.head,
        'post': session.post,
    }[method]
    req = method(url, stream=stream, headers=headers, timeout=timeout, verify=verify_ssl, **kwargs)
    req.raise_for_status()
    return req

def sanitize_filename(text, exclusions=''):
    to_remove = FILENAME_BADCHARS
    for exclude in exclusions:
        to_remove = to_remove.replace(exclude, '')

    for char in to_remove:
        text = text.replace(char, '')

    (drive, path) = os.path.splitdrive(text)
    path = path.replace(':', '')
    text = drive + path

    return text

def sanitize_url(url):
    url = url.replace('%3A//', '://')
    return url

def touch(filename):
    f = open(filename, 'ab')
    f.close()
    return


def download_argparse(args):
    url = args.url

    url = clipext.resolve(url)
    callback = {
        None: Progress1,
        '1': Progress1,
        '2': Progress2,
    }.get(args.callback, args.callback)

    bytespersecond = args.bytespersecond
    if bytespersecond is not None:
        bytespersecond = bytestring.parsebytes(bytespersecond)

    headers = {}
    if args.range is not None:
        headers['range'] = 'bytes=%s' % args.range

    retry = args.retry
    if not retry:
        retry = 1

    while retry != 0:
        # Negative numbers permit infinite retries.
        try:
            download_file(
                url=url,
                localname=args.localname,
                bytespersecond=bytespersecond,
                callback_progress=callback,
                do_head=args.no_head is False,
                headers=headers,
                overwrite=args.overwrite,
                timeout=args.timeout,
                verbose=True,
                verify_ssl=args.no_ssl is False,
            )
        except (NotEnoughBytes, requests.exceptions.ConnectionError, requests.exceptions.ChunkedEncodingError):
            retry -= 1
            if retry == 0:
                raise
        else:
            break


def main(argv):
    parser = argparse.ArgumentParser()

    parser.add_argument('url')
    parser.add_argument('localname', nargs='?', default=None)
    parser.add_argument('-c', '--callback', dest='callback', default=Progress1)
    parser.add_argument('-bps', '--bytespersecond', dest='bytespersecond', default=None)
    parser.add_argument('-ow', '--overwrite', dest='overwrite', action='store_true')
    parser.add_argument('-r', '--range', dest='range', default=None)
    parser.add_argument('--timeout', dest='timeout', type=int, default=TIMEOUT)
    parser.add_argument('--retry', dest='retry', nargs='?', type=int, default=1)
    parser.add_argument('--no-head', dest='no_head', action='store_true')
    parser.add_argument('--no-ssl', dest='no_ssl', action='store_true')
    parser.set_defaults(func=download_argparse)

    args = parser.parse_args(argv)
    return args.func(args)

if __name__ == '__main__':
    raise SystemExit(main(sys.argv[1:]))



'''
This function is slow and ugly, but I need a way to safely print unicode strings
on systems that don't support it without crippling those who do.
'''
import sys

def safeprint_handle(text, *, file_handle, end='\n'):
    for character in text:
        try:
            file_handle.write(character)
        except UnicodeError:
            file_handle.write('?')
    file_handle.write(end)

def safeprint_stdout(text, *, end='\n'):
    safeprint_handle(text=text, file_handle=sys.stdout, end=end)
    sys.stdout.flush()

def safeprint(text, *, file_handle=None, end='\n'):
    if file_handle is not None:
        return safeprint_handle(text=text, file_handle=file_handle, end=end)
    else:
        return safeprint_stdout(text=text, end=end)



import math
import sys

def hms_to_seconds(hms):
    '''
    Convert hh:mm:ss string to an integer seconds.
    '''
    parts = hms.split(':')
    seconds = 0
    if len(parts) > 3:
        raise ValueError(f'{hms} doesn\'t match the HH:MM:SS format.')
    if len(parts) == 3:
        seconds += int(parts[0]) * 3600
        parts.pop(0)
    if len(parts) == 2:
        seconds += int(parts[0]) * 60
        parts.pop(0)
    if len(parts) == 1:
        seconds += float(parts[0])
    return seconds

def seconds_to_hms(seconds):
    '''
    Convert integer number of seconds to an hh:mm:ss string.
    Only the necessary fields are used.
    '''
    seconds = math.ceil(seconds)
    (minutes, seconds) = divmod(seconds, 60)
    (hours, minutes) = divmod(minutes, 60)
    parts = []
    if hours:
        parts.append(hours)
    if hours or minutes:
        parts.append(minutes)
    parts.append(seconds)
    hms = ':'.join(f'{part:02d}' for part in parts)
    return hms

def main(args):
    from voussoirkit import pipeable
    for line in pipeable.go(args, strip=True, skip_blank=True):
        if ':' in line:
            line = hms_to_seconds(line)
        else:
            line = float(line)
            if line > 60:
                line = seconds_to_hms(line)

        pipeable.output(f'{line}')

if __name__ == '__main__':
    raise SystemExit(main(sys.argv[1:]))



'''
The Backoff classes are intended to be used to control `time.sleep` with
varying backoff strategies.

Example:

bo = backoff.Linear(m=1, b=1, max=30)
while True:
    try:
        something()
    except Exception:
        time.sleep(bo.next())
    else:
        bo.reset()

If you want to add random fuzziness to your sleeps, that should be done on the
calling end. For example, `bo.next() + (random.random() - 0.5)`.
'''
class Backoff:
    def __init__(self, max):
        if max is None:
            pass
        elif max <= 0:
            raise ValueError(f'max must be positive, not {max}.')
        self.max = max

    def next(self):
        y = self._calc()
        if self.max is not None:
            y = min(y, self.max)
        self.x += 1
        return y

    def rewind(self, steps):
        self.x = max(0, self.x - steps)

    def reset(self):
        self.x = 0

####################################################################################################

class Exponential(Backoff):
    '''
    Exponential backoff produces next = (a**x) + b.
    '''
    def __init__(self, a, b, *, max):
        super().__init__(max)
        self.x = 0
        self.a = a
        self.b = b
        self.max = max

    def _calc(self):
        return (self.a ** self.x) + self.b

class Linear(Backoff):
    '''
    Linear backoff produces next = (m * x) + b.
    '''
    def __init__(self, m, b, *, max):
        super().__init__(max)
        self.x = 0
        self.m = m
        self.b = b
        self.max = max

    def _calc(self):
        return (self.m * self.x) + self.b

class Quadratic(Backoff):
    '''
    Quadratic backoff produces next = (a * x**2) + (b * x) + c.
    '''
    def __init__(self, a, b, c, *, max):
        super().__init__(max)
        self.x = 0
        self.a = a
        self.b = b
        self.c = c
        self.max = max

    def _calc(self):
        return (self.a * self.x**2) + (self.b * self.x) + self.c

'''
people are backing off
you'll get used to it
'''



class Sentinel:
    '''
    Sentinel objects are used when you need to have some kind of default,
    placeholder, or special value that can't be confused with any other value
    in your program. For example, if you are waiting for a function to return a
    value, but you use `None` as a placeholder, someone might get confused and
    think that the function actually returned None.

    You can get cheap sentinels by just creating plain Python `object()`s, but
    they are bad for printing because you'll just see an ID and not know what
    the sentinel represents. This simple Sentinel class lets you give them a
    name and adjust their truthyness which can be useful.

    Some implementations of sentinels also make them singletons. These ones are
    not singletons! Separate sentinels will never == each other even if you use
    the same name!
    '''
    def __init__(self, name, truthyness=True):
        self.name = name
        self.truthyness = truthyness

    def __bool__(self):
        return bool(self.truthyness)

    def __repr__(self):
        return f'<Sentinel {repr(self.name)} id={id(self)} like {bool(self.truthyness)}>'



import collections

class Cache:
    def __init__(self, maxlen):
        self.maxlen = maxlen
        self.cache = collections.OrderedDict()

    def __contains__(self, key):
        return key in self.cache

    def __getitem__(self, key):
        value = self.cache.pop(key)
        self.cache[key] = value
        return value

    def __len__(self):
        return len(self.cache)

    def __setitem__(self, key, value):
        try:
            self.cache.pop(key)
        except KeyError:
            if len(self.cache) >= self.maxlen:
                self.cache.popitem(last=False)
        self.cache[key] = value

    def clear(self):
        self.cache.clear()

    def get(self, key, fallback=None):
        try:
            return self[key]
        except KeyError:
            return fallback

    def pop(self, key):
        return self.cache.pop(key)

    def remove(self, key):
        try:
            self.pop(key)
        except KeyError:
            pass



'''
The documentation for the classes and methods are below. Here are some examples
of threadpool in use:

1. Powering a single api scraping generator with many threads:

>>> pool = threadpool.ThreadPool(thread_count, paused=True)
>>> job_gen = ({'function': api.get_item, 'kwargs': {'id': i}} for i in range(lower, upper+1))
>>> pool.add_generator(job_gen)
>>> for job in pool.result_generator():
>>>     if job.exception:
>>>         raise job.exception
>>>     if job.value is not None:
>>>         yield job.value

2. Git-fetching a bunch of repositories with no error handling:

>>> def git_fetch(d):
>>>     command = [GIT, '-C', d, 'fetch', '--all']
>>>     print(command)
>>>     subprocess.check_output(command, stderr=subprocess.STDOUT)
>>>
>>> def callback(job):
>>>     if job.exception:
>>>         print(f'{job.name} caused {job.exception}.')
>>>
>>> pool = threadpool.ThreadPool(thread_count, paused=False)
>>> kwargss = [{'function': git_fetch, 'args': [d], 'name': d, 'callback': callback} for d in dirs]
>>> pool.add_many(kwargss)
>>> pool.join()
'''
import collections
import logging
import queue
import threading
import traceback

from voussoirkit import lazychain
from voussoirkit import sentinel

log = logging.getLogger('threadpool')

PENDING = sentinel.Sentinel('PENDING')
RUNNING = sentinel.Sentinel('RUNNING')
FINISHED = sentinel.Sentinel('FINISHED')
RAISED = sentinel.Sentinel('RAISED')

NO_MORE_JOBS = sentinel.Sentinel('NO_MORE_JOBS')

NO_RETURN = sentinel.Sentinel('NO_RETURN', truthyness=False)
NO_EXCEPTION = sentinel.Sentinel('NO_EXCEPTION', truthyness=False)

class ThreadPoolException(Exception):
    pass

class PoolClosed(ThreadPoolException):
    pass

class PooledThread:
    def __init__(self, pool):
        self.pool = pool
        self.thread = threading.Thread(target=self.start)
        self.thread.daemon = True
        self.thread.start()

    def __repr__(self):
        return f'PooledThread {self.thread}'

    def _run_once(self):
        # Any exceptions caused by the job's primary function are already
        # wrapped safely, but there are two other sources of potential
        # exceptions:
        # 1. A generator given to add_generator that encounters an exception
        #    while generating the kwargs causes get_next_job to raise.
        # 2. The callback function given to the Job raises.
        # It's hard to say what the correct course of action is, but I
        # realllly don't want them taking down the whole worker thread.
        try:
            job = self.pool.get_next_job()
        except BaseException:
            traceback.print_traceback()
            return

        if job is NO_MORE_JOBS:
            return NO_MORE_JOBS

        log.debug('%s is running job %s.', self, job)
        self.pool._running_count += 1
        try:
            job.run()
        except BaseException:
            traceback.print_traceback()
        self.pool._running_count -= 1

    def join(self):
        log.debug('%s is joining.', self)
        self.thread.join()

    def start(self):
        while True:
            # Let's wait for jobs_available first and unpaused second.
            # If the time between the two waits is very long, the worst thing
            # that can happen is there are no more jobs by the time we get
            # there, and the loop comes around again. On the other hand, if
            # unpaused.wait is first and the time until available.wait is very
            # long, we might wind up running a job despite the user pausing
            # the pool in the interim.
            self.pool._jobs_available.wait()
            self.pool._unpaused_event.wait()
            status = self._run_once()
            if status is NO_MORE_JOBS and self.pool.closed:
                break

class ThreadPool:
    '''
    The ThreadPool is used to perform large numbers of tasks using a pool of
    worker threads. Jobs are run in the order they are added.

    The pool supports two main paradigms of usage:

    1. Callback / async style
        If the job function performs your desired side effects by itself, or is
        given a callback function, you can simply add it to the pool and wait
        for it to run.

    2. Generator style
        If you want to yield the job results back to the main thread for
        processing (e.g. you are feeding the results into sqlite, which must be
        done on the thread which opened the sqlite connection), you can use
        `result_generator` to get each job in the order they were added to the
        pool. This style also makes it easier to terminate the main thread when
        a single job encounters an issue. Just `raise job.exception`.
    '''
    def __init__(self, size, paused=True):
        '''
        size:
            The number of worker threads.

        paused:
            If True, the pool will start in a paused state and you will have to
            call `start` to start it. If False, the pool will run as soon as
            jobs are added to it.
        '''
        if not isinstance(size, int):
            raise TypeError(f'size must be an int, not {type(size)}.')
        if size < 1:
            raise ValueError(f'size must be >= 1, not {size}.')

        self._unpaused_event = threading.Event()
        if not paused:
            self._unpaused_event.set()

        self._jobs_available = threading.Event()

        self._closed = False
        self._running_count = 0
        self._result_queue = None
        self._pending_jobs = lazychain.LazyChain()
        self._job_manager_lock = threading.Lock()

        self._size = size
        self._threads = [PooledThread(pool=self) for x in range(size)]

    @property
    def closed(self):
        return self._closed

    @property
    def paused(self):
        return not self._unpaused_event.is_set()

    @property
    def running_count(self):
        return self._running_count

    @property
    def size(self):
        return self._size

    def assert_not_closed(self):
        '''
        If the pool is closed (because you called `join`), raise PoolClosed.
        Otherwise do nothing.
        '''
        if self._closed:
            raise PoolClosed()

    def add(self, function, *, name=None, callback=None, args=tuple(), kwargs=dict()):
        '''
        Add a new job to the pool.

        See the Job class for parameter details.
        '''
        self.assert_not_closed()

        job = Job(
            pool=self,
            function=function,
            name=name,
            args=args,
            kwargs=kwargs,
        )
        self._pending_jobs.append(job)
        self._jobs_available.set()

        return job

    def add_generator(self, kwargs_gen):
        '''
        Add jobs from a generator which yields kwarg dictionaries. Unlike
        `add` and `add_many`, the Job objects are not returned by this method
        (since they don't exist yet!). If you want them, use `result_generator`
        to iterate the pool's jobs as they complete. Otherwise, they should
        have their own side effects or use a callback.

        See the Job class for kwarg details.
        '''
        self.assert_not_closed()

        these_jobs = (Job(pool=self, **kwargs) for kwargs in kwargs_gen)
        self._pending_jobs.extend(these_jobs)
        self._jobs_available.set()

    def add_many(self, kwargss):
        '''
        Add multiple new jobs to the pool at once. This is better than calling
        `add` in a loop because we only have to aquire the lock one time.

        Provide an iterable of kwarg dictionaries. That is:
        [
            {'function': print, 'args': [4], 'name': '4'},
            {'function': sample, 'kwargs': {'x': 2}},
        ]

        See the Job class for kwarg details.
        '''
        self.assert_not_closed()

        kwargss = list(kwargss)
        if not kwargss:
            raise ValueError(f'{kwargss} must not be empty.')

        these_jobs = [Job(pool=self, **kwargs) for kwargs in kwargss]
        self._pending_jobs.extend(these_jobs)
        self._jobs_available.set()

        return these_jobs

    def get_next_job(self):
        with self._job_manager_lock:
            try:
                job = next(self._pending_jobs)
            except StopIteration:
                # If we ARE closed, we want to keep the flag set so that all
                # the threads can keep waking up and seeing no more jobs.
                if not self.closed:
                    self._jobs_available.clear()
                return NO_MORE_JOBS

            if self._result_queue is not None:
                # This will block if the queue is full.
                self._result_queue.put(job)

            return job

    def join(self):
        '''
        Permanently close the pool, preventing any new jobs from being added,
        and block until all jobs are complete.
        '''
        log.debug('%s is joining.', self)
        self._closed = True
        # The threads which are currently paused at _jobs_available.wait() need
        # to be woken up so they can realize the pool is closed and break.
        self._jobs_available.set()
        self.start()
        for thread in self._threads:
            thread.join()

    def result_generator(self, *, buffer_size=None):
        '''
        This generator will start the job pool, then yield finished/raised Job
        objects in the order they were added. Note that a slow job will
        therefore hold up the generator, though it will not stop the job pool
        from running and spawning new jobs in their other threads.

        For best results, you should create the pool in the paused state, add
        your jobs, then use this method to start the pool. Any jobs that run
        while the result_generator is not active will not be stored, since we
        don't necessarily know if this method will ever be used. So, any jobs
        that start before the result_generator is active will not be yielded
        and will simply be lost to garbage collection.

        If more jobs are added while the generator is running, they will be
        yielded as expected.

        When there are no more outstanding jobs, the generator will stop
        iteration and return. If the pool was paused before generating, it
        will be paused again. This prevents subsequently added jobs from being
        lost as described.

        buffer_size:
            The size of the buffer which holds jobs before they are yielded.
            If you expect your production to outpace your consumption, you may
            wish to set this value to prevent high memory usage. When the buffer
            is full, new jobs will be blocked from starting.
        '''
        if self._result_queue is not None:
            raise TypeError('The result generator is already open.')

        self._result_queue = queue.Queue(maxsize=buffer_size or 0)

        was_paused = self.paused

        self.start()
        # Considerations for the while loop condition:
        # Why `jobs_available.is_set`: Consider a group of slow-running threads
        # are launched and the jobs are added to the result_queue. The caller
        # of this generator consumes all of them before the threads finish and
        # start a new job. So, we need to watch jobs_available.is_set to know
        # that even though the result_queue is currently empty, we can expect
        # more to be ready soon and shouldn't break yet.
        # Why `not results_queue.empty`: Consider a group of fast-running
        # threads are launched, and exhaust all available jobs. So, we need to
        # watch that result_queue is not empty and has more results.
        # Why not `not closed`: After the pool is closed, the outstanding jobs
        # still need to finish. Closing does not imply pausing or cancelling
        # jobs.
        while self._jobs_available.is_set() or not self._result_queue.empty():
            job = self._result_queue.get()
            job.join()
            yield job
            self._result_queue.task_done()
        self._result_queue = None

        if was_paused:
            self.pause()

    def pause(self):
        self._unpaused_event.clear()

    def start(self):
        self._unpaused_event.set()

class Job:
    '''
    Each job contains one function that it will call when it is started.

    If the function completes successfully (status is threadpool.FINISHED) you
    will find the return value in `job.value`. If it raises an exception
    (status is threadpool.RAISED), you'll find it in `job.exception`, although
    the thread itself will not raise.

    All job threads are daemons and will not prevent the main thread from
    terminating. Call `job.join()` or `pool.join()` in the main thread to
    ensure jobs complete.
    '''
    def __init__(self, pool, function, *, name=None, callback=None, args=tuple(), kwargs=dict()):
        '''
        When this job is started, `function(*args, **kwargs)` will be called.

        name:
            An optional value that will appear in the repr of the job and
            has no other purpose. Use this if you intend to print(job) and want
            a human friendly name string.

        callback:
            An optional function which will be called as `callback(job)` after
            the job is finished running. Use this for async-style processing of
            the job. Note that the callback is called via the job's thread, so
            make sure it is memory safe.
        '''
        self.pool = pool
        self.name = name
        self.status = PENDING
        self.function = function
        self.callback = callback
        self.args = args
        self.kwargs = kwargs
        self.value = NO_RETURN
        self.exception = NO_EXCEPTION

        self._done_event = threading.Event()

    def __repr__(self):
        if self.name:
            return f'<{self.status.name} Job {repr(self.name)}>'
        else:
            return f'<{self.status.name} Job on {self.function}>'

    def run(self):
        self.status = RUNNING
        try:
            self.value = self.function(*self.args, **self.kwargs)
            self.status = FINISHED
        except BaseException as exc:
            self.exception = exc
            self.status = RAISED

        if self.callback is not None:
            self.callback(self)

        self._done_event.set()

    def join(self):
        '''
        Block until this job runs and completes.
        '''
        self._done_event.wait()



'''
Instead of adding every program's directory to my system PATH, I prefer having
a single directory on my PATH which contains Windows shortcuts (lnk files)
pointing to each program.
Windows lnks are different from softlinks because they maintain the program's
real directory necessary for loading nearby dlls etc.
However, this breaks `shutil.which` --> `subprocess.run` because subprocess.run
with shell=False does not interpret lnks, and needs a direct path to the exe.

So, this module provides a function `which` that if it finds an lnk file, will
return the exe path, otherwise behaves the same as normal shutil which.
'''
import os
import shutil
try:
    import winshell
except ImportError:
    winshell = None

def which(cmd, *args, **kwargs):
    path = shutil.which(cmd, *args, **kwargs)

    if path is None:
        return None

    if os.name == 'nt' and winshell and path.lower().endswith('.lnk'):
        path = winshell.Shortcut(path).path

    return path



ESCAPE_SEQUENCES = {
    '\\': '\\',
    '"': '"',
}

BINARY_OPERATORS = {'AND', 'OR', 'XOR'}
UNARY_OPERATORS = {'NOT'}
PRECEDENCE = ['NOT', 'AND', 'XOR', 'OR']
OPERATORS = BINARY_OPERATORS | UNARY_OPERATORS

# Sentinel values used for breaking up the tokens, so we dont' have to use
# strings '(' and ')' which can get confused with user input.
PAREN_OPEN = object()
PAREN_CLOSE = object()

DEFAULT_MATCH_FUNCTION = str.__contains__

MESSAGE_WRITE_YOUR_OWN_MATCHER = '''
The default match function is {function}.
Consider passing your own `match_function`, which accepts two
positional arguments:
1. The object being tested.
2. The Expression token, a string.
'''.strip()

def func_and(values):
    return all(values)

def func_or(values):
    return any(values)

def func_xor(values):
    values = list(values)
    return values.count(True) % 2 == 1

def func_not(value):
    value = list(value)
    if len(value) != 1:
        raise ValueError('NOT only takes 1 value')
    return not value[0]

OPERATOR_FUNCTIONS = {
    'AND': func_and,
    'OR': func_or,
    'XOR': func_xor,
    'NOT': func_not,
}

class NoTokens(Exception):
    pass

class ExpressionTree:
    def __init__(self, token, parent=None):
        self.children = []
        self.parent = parent
        self.token = token

    def __str__(self):
        if self.token is None:
            return '""'

        self_token = str(self.token)

        if self_token not in OPERATORS:
            t = self_token
            t = t.replace('"', '\\"')
            t = t.replace('(', '\\(')
            t = t.replace(')', '\\)')
            if ' ' in t:
                t = f'"{t}"'
            return t

        if len(self.children) == 1:
            child = self.children[0]
            childstring = str(child)
            if child.token in OPERATORS:
                return f'{self_token}({childstring})'
            return f'{self_token} {childstring}'

        children = []
        for child in self.children:
            childstring = str(child)
            if child.token in OPERATORS:
                childstring = f'({childstring})'
            children.append(childstring)

        if len(children) == 1:
            return f'{self_token} {children[0]}'

        s = f' {self_token} '
        s = s.join(children)
        return s

    @classmethod
    def parse(cls, tokens, spaces=0):
        if isinstance(tokens, str):
            tokens = tokenize(tokens)

        if tokens == []:
            raise NoTokens()

        if isinstance(tokens[0], list):
            current = cls.parse(tokens[0], spaces=spaces+1)
        else:
            current = cls(token=tokens[0])

        for token in tokens[1:]:
            if isinstance(token, list):
                new = cls.parse(token, spaces=spaces+1)
            else:
                new = cls(token=token)

            if 0 == 1:
                pass

            elif current.token not in OPERATORS:
                if new.token in BINARY_OPERATORS:
                    if len(new.children) == 0:
                        new.children.append(current)
                        current.parent = new
                        current = new
                else:
                    raise Exception(f'Expected binary operator, got {new.token}.')

            elif current.token in BINARY_OPERATORS:
                if new.token in BINARY_OPERATORS:
                    if new.token == current.token:
                        for child in new.children:
                            child.parent = current
                        current.children.extend(new.children)
                    else:
                        if len(new.children) == 0:
                            new.children.append(current)
                            current.parent = new
                            current = new
                        else:
                            current.children.append(new)
                            new.parent = current

                elif new.token in UNARY_OPERATORS:
                    if len(new.children) == 0:
                        current.children.append(new)
                        new.parent = current
                        current = new
                    else:
                        current.children.append(new)
                        new.parent = current

                elif new.token not in OPERATORS:
                    if len(current.children) > 0:
                        current.children.append(new)
                        new.parent = current
                    else:
                        raise Exception('Expected current children > 0.')

            elif current.token in UNARY_OPERATORS:
                if len(current.children) == 0:
                    current.children.append(new)
                    new.parent = current
                    if current.parent is not None:
                        current = current.parent
                elif new.token in BINARY_OPERATORS:
                    if len(new.children) == 0:
                        new.children.append(current)
                        current.parent = new
                        current = new
                    else:
                        current.children.append(new)
                        new.parent = current
                        if current.parent is not None:
                            current = current.parent
                else:
                    raise Exception('Expected new to be my operand or parent binary.')

        current = current.rootmost()
        return current

    def _evaluate(self, text, match_function=None):
        if self.token not in OPERATORS:
            if match_function is None:
                match_function = DEFAULT_MATCH_FUNCTION

            value = match_function(text, self.token)
            return value

        operator_function = OPERATOR_FUNCTIONS[self.token]
        children = (child.evaluate(text, match_function=match_function) for child in self.children)
        return operator_function(children)

    def diagram(self):
        if self.token is None:
            return '""'
        t = self.token
        if ' ' in t:
            t = f'"{t}"'

        output = t
        indent = 1
        for child in self.children:
            child = child.diagram()
            for line in child.splitlines():
                output += (' ' * indent)
                output += line + '\n'
                indent = len(t) + 1
        output = output.strip()

        return output

    def evaluate(self, text, match_function=None):
        if match_function is None:
            match_function = DEFAULT_MATCH_FUNCTION

        try:
            return self._evaluate(text, match_function)
        except Exception as e:
            if match_function is DEFAULT_MATCH_FUNCTION:
                message = MESSAGE_WRITE_YOUR_OWN_MATCHER.format(function=DEFAULT_MATCH_FUNCTION)
                override = Exception(message)
                raise override from e
            raise e

    @property
    def is_leaf(self):
        return self.token not in OPERATORS

    def map(self, function):
        '''
        Apply this function to all of the operands.
        '''
        for node in self.walk_leaves():
            node.token = function(node.token)

    def prune(self):
        '''
        Remove any nodes where `token` is None.
        '''
        self.children = [child for child in self.children if child.token is not None]

        for child in self.children:
            child.prune()

        if self.token in OPERATORS and len(self.children) == 0:
            self.token = None
            if self.parent is not None:
                self.parent.children.remove(self)

    def rootmost(self):
        current = self
        while current.parent is not None:
            current = current.parent
        return current

    def walk(self):
        yield self
        for child in self.children:
            yield from child.walk()

    def walk_leaves(self):
        for node in self.walk():
            if node.is_leaf:
                yield node


def implied_tokens(tokens):
    '''
    1. If two operands are directly next to each other, or an operand is followed
        by a unary operator, it is implied that there is an AND between them.
        '1 2' -> '1 AND 2'
        '1 NOT 2' -> '1 AND NOT 2'

    2. If an expression begins or ends with an invalid operator, remove it.
        'AND 2' -> '2'
        '2 AND' -> '2'

    3. If a parenthetical term contains only 1 item, the parentheses can be removed.
        '(a)' -> 'a'
        '(NOT a)' -> 'NOT a'
        '(a OR)' -> '(a)' (by rule 2) -> 'a'

    4. If two operators are next to each other, except for binary-unary,
        keep only the first.
        '1 OR AND 2' -> '1 OR 2'
        '1 NOT AND 2' -> '1 AND NOT AND 2' (by rule 1) -> '1 AND NOT 2'
        'NOT NOT 1' -> 'NOT 1'
        '1 AND NOT NOT 2' -> '1 AND NOT 2'
    '''
    final_tokens = []
    has_operand = False
    has_binary_operator = False
    has_unary_operator = False

    if len(tokens) == 1 and not isinstance(tokens[0], str):
        # [['A' 'AND' 'B']] -> ['A' 'AND' 'B']
        tokens = tokens[0]

    for token in tokens:
        skip_this = False
        while isinstance(token, (list, tuple)):
            if len(token) == 0:
                # Delete empty parentheses.
                skip_this = True
                break
            if len(token) == 1:
                # Take singular terms out of their parentheses.
                token = token[0]
            else:
                previous = token
                token = implied_tokens(token)
                if previous == token:
                    break

        if skip_this:
            continue

        if isinstance(token, str) and token in OPERATORS:
            this_binary = token in BINARY_OPERATORS
            this_unary = not this_binary

            # 'NOT AND' and 'AND AND' are malformed...
            if this_binary and (has_binary_operator or has_unary_operator):
                continue
            # ...'NOT NOT' is malformed...
            if this_unary and has_unary_operator:
                continue
            # ...but AND NOT is okay.

            # 'AND test' is malformed
            if this_binary and not has_operand:
                continue

            if this_unary and has_operand:
                final_tokens.append('AND')

            has_unary_operator = this_unary
            has_binary_operator = this_binary
            has_operand = False

        else:
            if has_operand:
                final_tokens.append('AND')
            has_unary_operator = False
            has_binary_operator = False
            has_operand = True

        final_tokens.append(token)

    if has_binary_operator or has_unary_operator:
        final_tokens.pop(-1)

    return final_tokens

def order_operations(tokens):
    for (index, token) in enumerate(tokens):
        if isinstance(token, list):
            tokens[index] = order_operations(token)

    if len(tokens) < 5:
        return tokens

    index = 0
    slice_start = None
    slice_end = None
    precedence_stack = []
    while index < len(tokens):
        token = tokens[index]
        try:
            precedence = PRECEDENCE.index(token)
        except ValueError:
            precedence = None

        if precedence is None:
            index += 1
            continue
        precedence_stack.append(precedence)

        if token in UNARY_OPERATORS:
            slice_start = index
            slice_end = index + 2

        elif len(precedence_stack) > 1:
            if precedence_stack[-1] < precedence_stack[-2]:
                slice_start = index - 1
                slice_end = None
            elif precedence_stack[-2] < precedence_stack[-1]:
                slice_end = index

        if slice_start is None or slice_end is None:
            index += 1
            continue

        tokens[slice_start:slice_end] = [tokens[slice_start:slice_end]]
        slice_start = None
        slice_end = None
        for x in range(2):
            if not precedence_stack:
                break

            delete = precedence_stack[-1]
            while precedence_stack and precedence_stack[-1] == delete:
                index -= 1
                precedence_stack.pop(-1)

        index += 1

    if slice_start is not None:
        slice_end = len(tokens)
        tokens[slice_start:slice_end] = [tokens[slice_start:slice_end]]

    return tokens

def sublist_tokens(tokens, _from_index=0, depth=0):
    '''
    Given a list of tokens, replace parentheses with actual sublists.
    ['1', 'AND', '(', '3', 'OR', '4', ')'] ->
    ['1', 'AND', ['3', 'OR', '4']]

    Unclosed parentheses are automatically closed at the end.
    '''
    final_tokens = []
    index = _from_index
    while index < len(tokens):
        token = tokens[index]
        index += 1
        if token is PAREN_OPEN:
            (token, index) = sublist_tokens(tokens, _from_index=index, depth=depth+1)
        if token is PAREN_CLOSE:
            break
        final_tokens.append(token)
    if _from_index == 0:
        return final_tokens
    else:
        return (final_tokens, index)

def tokenize(expression):
    '''
    Break the string into a list of  tokens. Spaces are the delimiter unless
    they are inside quotation marks.

    Quotation marks and parentheses can be escaped by preceeding with a backslash '\\'

    Opening and closing parentheses are put into their own token unless
    escaped / quoted.

    Extraneous closing parentheses are ignored completely.

    '1 AND(4 OR "5 6") OR \\(test\\)' ->
    ['1', 'AND', '(', '4', 'OR', '5 6', ')', 'OR', '\\(test\\)']
    '''
    current_word = []
    in_escape = False
    in_quotes = False
    paren_depth = 0
    tokens = []
    for character in expression:
        if in_escape:
            in_escape = False

        elif character in {'(', ')'} and not in_quotes:
            if character == '(':
                sentinel = PAREN_OPEN
                paren_depth += 1
            elif character == ')':
                sentinel = PAREN_CLOSE
                paren_depth -= 1

            if paren_depth >= 0:
                tokens.append(''.join(current_word))
                tokens.append(sentinel)
                current_word.clear()
                continue
            else:
                continue

        elif character == '\\':
            in_escape = True
            continue

        elif character == '"':
            in_quotes = not in_quotes
            continue

        elif character.isspace() and not in_quotes:
            tokens.append(''.join(current_word))
            current_word.clear()
            continue

        current_word.append(character)

    tokens.append(''.join(current_word))
    tokens = [w for w in tokens if w != '']
    tokens = sublist_tokens(tokens)
    tokens = implied_tokens(tokens)
    tokens = order_operations(tokens)
    return tokens

if __name__ == '__main__':
    tests = [
        '[sci-fi] OR [pg-13]',
        '([sci-fi] OR [war]) AND [r]',
        '[r] XOR [sci-fi]',
        '"mark hamill" "harrison ford"',
    ]
    teststrings = {
        'Star Wars': '[harrison ford] [george lucas] [sci-fi] [pg] [carrie fisher] [mark hamill] [space]',
        'Blade Runner': '[harrison ford] [ridley scott] [neo-noir] [dystopian] [sci-fi] [r]',
        'Indiana Jones': '[harrison ford] [steven spielberg] [adventure] [pg-13]',
        'Apocalypse Now': '[harrison ford] [francis coppola] [r] [war] [drama]'
    }
    for token in tests:
        print('start:', token)
        token = tokenize(token)
        print('implied:', token)
        e = ExpressionTree.parse(token)
        print('tree:', e)
        for (name, teststring) in teststrings.items():
            print('Matches', name, ':', e.evaluate(teststring))
        print()



'''
Overall, I like types.SimpleNamespace. But I decided to make dotdict because:
1. SimpleNamespace is a cumbersome name to type and to look at.
2. I wanted my class to support default values.
'''
from voussoirkit import sentinel

NO_DEFAULT = sentinel.Sentinel('NO_DEFAULT')

class DotDict:
    def __init__(self, __dict=None, *, default=NO_DEFAULT, **kwargs):
        self.__default = default
        if __dict:
            self.__dict__.update(__dict)
        self.__dict__.update(**kwargs)

    def __getattr__(self, key):
        try:
            return self.__dict__[key]
        except KeyError:
            if self.__default is not NO_DEFAULT:
                return self.__default
            raise

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __repr__(self):
        display = self.__dict__.copy()
        display.pop('_DotDict__default')
        return f'DotDict {display}'



from logging import *

_getLogger = getLogger

LOUD = 1

def getLogger(*args, **kwargs):
    log = _getLogger(*args, **kwargs)
    add_loud(log)
    return log

def add_loud(log):
    addLevelName(LOUD, 'LOUD')
    log.loud = lambda *args, **kwargs: log.log(LOUD, *args, **kwargs)

def set_level_by_argv(log, argv):
    basicConfig()
    argv = argv[:]

    if '--loud' in argv:
        log.setLevel(LOUD)
        argv.remove('--loud')
    elif '--debug' in argv:
        log.setLevel(DEBUG)
        argv.remove('--debug')
    elif '--quiet' in argv:
        log.setLevel(ERROR)
        argv.remove('--quiet')
    else:
        log.setLevel(INFO)

    return argv



'''
Find time, filesize, or bitrate, given two of the three.

For example:

kbps.py --time 1:00:00 --size 2g
kbps.py --time 1:00:00 --kbps 4660
kbps.py --size 2g --kpbps 4660
'''
import argparse
import sys

from voussoirkit import bytestring
from voussoirkit import hms

def kbps(time=None, size=None, kbps=None):
    if [time, size, kbps].count(None) != 1:
        raise ValueError('Incorrect number of unknowns.')

    if time is None:
        kilobits = size / 128
        time = kilobits / kbps

    if size is None:
        kibs = kbps / 8
        size = kibs * 1024
        size *= time

    if kbps is None:
        kibs = size / 1024
        kilobits = kibs * 8
        kbps = kilobits / time

    return (time, size, kbps)

def kbps_argparse(args):
    time = args.time and hms.hms_to_seconds(args.time)
    size = args.size and bytestring.parsebytes(args.size)
    kibs = args.kbps and int(args.kbps)

    (time, size, kibs) = kbps(time=time, size=size, kbps=kibs)

    print(f'{hms.seconds_to_hms(time)} @ {kibs} kbps = {bytestring.bytestring(size)}')

def main(argv):
    parser = argparse.ArgumentParser()

    parser.add_argument('-t', '--time', dest='time', default=None)
    parser.add_argument('-s', '--size', dest='size', default=None)
    parser.add_argument('-k', '--kbps', dest='kbps', default=None)
    parser.set_defaults(func=kbps_argparse)

    args = parser.parse_args(argv)
    return args.func(args)

if __name__ == '__main__':
    raise SystemExit(main(sys.argv[1:]))



'''
This module provides functions for getting information about logical drives
on Windows.
'''
import ctypes
import re
import string
kernel32 = ctypes.windll.kernel32

def get_all_volumes():
    '''
    Return a list of volume paths like \\?\Volume{GUID}\ for all volumes,
    whether they are mounted or not.

    Note: This will include recovery / EFI partitions, which may not be what
    you're looking for. Also see get_drive_letters and get_drive_mounts.

    Thank you Duncan.
    https://stackoverflow.com/a/3075879
    '''
    volumes = []
    buf = ctypes.create_unicode_buffer(1024)
    length = ctypes.c_int32()
    handle = kernel32.FindFirstVolumeW(buf, ctypes.sizeof(buf))
    if handle:
        volumes.append(buf.value)
        while kernel32.FindNextVolumeW(handle, buf, ctypes.sizeof(buf)):
            volumes.append(buf.value)
        kernel32.FindVolumeClose(handle)
    return volumes

def get_drive_info(path):
    '''
    Given a drive path as either:
    - a letter like C or C:\, or
    - a mount path, or
    - a volume path like \\?\Volume{GUID},
    return a dictionary containing its attributes.

    Thanks Nicholas Orlowski
    http://stackoverflow.com/a/12056414
    '''
    letter_match = re.match(r'^([A-Z])(|:|:\\)$', path)
    if letter_match:
        letter = letter_match.group(1)
        path = f'{letter}:\\'

    if path.startswith('\\\\?\\Volume{'):
        mount = get_volume_mount(path)
    else:
        mount = path

    name_buffer = ctypes.create_unicode_buffer(1024)
    filesystem_buffer = ctypes.create_unicode_buffer(1024)
    serial_number = None
    max_component_length = None
    file_system_flags = None

    drive_active = kernel32.GetVolumeInformationW(
        ctypes.c_wchar_p(path),
        name_buffer,
        ctypes.sizeof(name_buffer),
        serial_number,
        max_component_length,
        file_system_flags,
        filesystem_buffer,
        ctypes.sizeof(filesystem_buffer),
    )
    info = {
        'active': bool(drive_active),
        'filesystem': filesystem_buffer.value,
        'name': name_buffer.value,
        'mount': mount,
    }
    return info

def get_drive_letters():
    '''
    Return a list of all connected drive letters as single-character strings.

    Drives which are mounted to paths instead of letters will not be returned.
    Use get_drive_mounts instead.

    Thanks RichieHindle
    http://stackoverflow.com/a/827398
    '''
    # "If the function succeeds, the return value is a bitmask representing the
    # currently available disk drives. Bit position 0 (the least-significant
    # bit) is drive A, bit position 1 is drive B, bit position 2 is drive C,
    # and so on."
    # https://docs.microsoft.com/en-us/windows/win32/api/fileapi/nf-fileapi-getlogicaldrives
    letters = []
    bitmask = ctypes.windll.kernel32.GetLogicalDrives()
    for letter in string.ascii_uppercase:
        if bitmask & 1:
            letters.append(letter)
        bitmask >>= 1
    return letters

def get_drive_map():
    '''
    Return a dict of {mount: info} for all connected drives, where mount is
    either a drive letter or mount path, and info is the dict returned
    by get_drive_info.
    '''
    drives = {mount: get_drive_info(mount) for mount in get_drive_mounts()}
    return drives

def get_drive_mounts():
    '''
    Return a list of all connected drives as either:
    - a letter like C:\ if the volume has a letter, or
    - a mount path if the volume does not have a letter.
    '''
    mounts = []
    for volume in get_all_volumes():
        mount = get_volume_mount(volume)
        if mount:
            mounts.append(mount)
    return mounts

def get_drive_mount_by_name(name):
    '''
    Given the name of a drive (the user-customizable name seen in Explorer),
    return the letter or mount path of that drive.

    Raises KeyError if it is not found.
    '''
    drives = get_drive_map()
    for (mount, info) in drives.items():
        if info['name'] == name:
            return mount
    raise KeyError(name)

def get_volume_mount(volume):
    '''
    Given a volume path like \\?\Volume{GUID}\, return either:
    - a letter like C:\ if the volume has a letter, or
    - a mount path if the volume does not have a letter, or
    - emptystring if the volume is not mounted at all.

    Thank you Duncan.
    https://stackoverflow.com/a/3075879

    Note: The API function is named "GetVolumePathNames..." in the plural,
    and Duncan's original answer uses .split('\0'), but in my testing the
    drives always contain only a single name.
    If the drive has a letter and mount path, only the letter is returned.
    If it has two mount paths, only the first one is returned.
    So, I'll just use a single return value until further notice.
    '''
    buf = ctypes.create_unicode_buffer(1024)
    length = ctypes.c_int32()
    kernel32.GetVolumePathNamesForVolumeNameW(
        ctypes.c_wchar_p(volume),
        buf,
        ctypes.sizeof(buf),
        ctypes.pointer(length),
    )
    return buf.value



'''
This module is designed to provide a GOOD ENOUGH means of identifying duplicate
files very quickly, so that more in-depth checks can be done on likely matches.
'''
import hashlib
import os
import sys

from voussoirkit import pathclass

SEEK_END = 2
CHUNK_SIZE = 2**20
FORMAT = '{size}_{hashtype}_{chunk_size}_{hash}'

HASH_CLASSES = {
    name: getattr(hashlib, name)
    for name in hashlib.algorithms_guaranteed
}

def equal_handle(handle1, handle2, *args, **kwargs):
    '''
    Given two handles, return True if they have the same quickid hash.
    '''
    size1 = handle1.seek(0, SEEK_END)
    size2 = handle2.seek(0, SEEK_END)
    handle1.seek(0)
    handle2.seek(0)
    if size1 != size2:
        return False

    id1 = quickid_handle(handle1, *args, **kwargs)
    id2 = quickid_handle(handle2, *args, **kwargs)
    return id1 == id2

def equal_file(filename1, filename2, *args, **kwargs):
    '''
    Given two files, return True if they have the same quickid hash.
    '''
    file1 = pathclass.Path(filename1)
    file2 = pathclass.Path(filename2)
    file1.assert_is_file()
    file2.assert_is_file()
    if file1.size != file2.size:
        return False
    with file1.open('rb') as handle1, file2.open('rb') as handle2:
        return equal_handle(handle1, handle2, *args, **kwargs)

def matches_handle(handle, other_id):
    '''
    Given a handle and a quickid hash, return True if the handle matches
    that hash.
    '''
    (other_size, hashtype, chunk_size, other_hash) = other_id.split('_')
    other_size = int(other_size)
    chunk_size = int(chunk_size)

    this_size = handle.seek(0, SEEK_END)
    handle.seek(0)
    if this_size != other_size:
        return False

    this_id = quickid_handle(handle, hashtype=hashtype, chunk_size=chunk_size)
    return this_id == other_id

def matches_file(filename, other_id):
    '''
    Given a file and a quickid hash, return True if the file matches
    that hash.
    '''
    file = pathclass.Path(filename)
    with file.open('rb') as handle:
        return matches_handle(handle, other_id)

def quickid_handle(handle, hashtype='md5', chunk_size=CHUNK_SIZE):
    '''
    Return the quickid hash for this handle.
    '''
    hasher = HASH_CLASSES[hashtype]()
    size = handle.seek(0, SEEK_END)
    handle.seek(0)

    if size <= 2 * chunk_size:
        hasher.update(handle.read())
    else:
        hasher.update(handle.read(chunk_size))
        handle.seek(-1 * chunk_size, SEEK_END)
        hasher.update(handle.read())

    output = FORMAT.format(
        size=size,
        hashtype=hashtype,
        chunk_size=chunk_size,
        hash=hasher.hexdigest(),
    )
    return output

def quickid_file(filename, *args, **kwargs):
    '''
    Return the quickid hash for this file.
    '''
    file = pathclass.Path(filename)
    file.assert_is_file()
    with file.open('rb') as handle:
        return quickid_handle(handle, *args, **kwargs)

def main(argv):
    print(quickid_file(argv[0]))

if __name__ == '__main__':
    raise SystemExit(main(sys.argv[1:]))
